@@ -75,9 +75,12 @@ sub parse_header {
     my $out = Compress::Snappy::decompress($data);
     $data = $out;
   } elsif ($encoding == SRL_PROTOCOL_ENCODING_SNAPPY_INCREMENTAL) {
-    die "Incremental Snappy encoding not implemented in hobodecoder. (yet. It's easy to do.)";
-  }
-  elsif ($encoding) {
+    print "Header says: Document body is Snappy-compressed (incremental).\n";
+    my $compressed_len = varint();
+    require Compress::Snappy;
+    my $out = Compress::Snappy::decompress($data);
+    $data = $out;
+  } elsif ($encoding) {
     die "Invalid encoding '" . ($encoding >> SRL_PROTOCOL_VERSION_BITS) . "'";
   }
 }
@@ -193,6 +196,12 @@ sub parse_sv {
     printf  "%6s  %2s %3s %s  Value:\n",("") x 3, $ind."  ";
     parse_sv($ind."    ");
   }
+  elsif ($o == SRL_HDR_OBJECTV_FREEZE) {
+    my $ofs= varint();
+    printf "%06u: %02x %03s %sOBJECTV_FREEZE(%d)\n", $p, $o, $bv, $ind, $ofs;
+    printf  "%6s  %2s %3s %s  Value:\n",("") x 3, $ind."  ";
+    parse_sv($ind."    ");
+  }
   elsif ($o == SRL_HDR_OBJECT) {
     printf "%06u: %02x %03s %sOBJECT\n", $p, $o, $bv, $ind;
     printf  "%6s  %2s %3s %s  Class:\n",("") x 3, $ind."  ";
@@ -200,6 +209,13 @@ sub parse_sv {
     printf  "%6s  %2s %3s %s  Value:\n",("") x 3, $ind."  ";
     parse_sv($ind."    ");
   }
+  elsif ($o == SRL_HDR_OBJECT_FREEZE) {
+    printf "%06u: %02x %03s %sOBJECT_FREEZE\n", $p, $o, $bv, $ind;
+    printf  "%6s  %2s %3s %s  Class:\n",("") x 3, $ind."  ";
+    parse_sv($ind."    ");
+    printf  "%6s  %2s %3s %s  Value:\n",("") x 3, $ind."  ";
+    parse_sv($ind."    ");
+  }
   elsif ($o == SRL_HDR_REGEXP) {
     printf "%06u: %02x %03s %sREGEXP\n", $p, $o, $bv, $ind;
     parse_sv($ind."  ");
